@@ -40,6 +40,7 @@ from .ParamTab import ParamTab
 from .PlotTab import PlotTab
 from .locopositioning_tab import LocoPositioningTab
 from .QualisysTab import QualisysTab
+from .LogClientTab import LogClientTab
 
 __author__ = 'Bitcraze AB'
 __all__ = []
@@ -56,4 +57,5 @@ available = [
     PlotTab,
     LocoPositioningTab,
     QualisysTab,
+    LogClientTab,
 ]
